from dataclasses import dataclass
from unittest.case import skipIf

from django.db.models import F
from django.test.testcases import TestCase
import mock

from ..records import Lambda, Adjunct, Callback
from django.test.utils import tag
from click.types import INT


try:
    from .celestials.models import Celestial
    from .celestials.galaxy import Stars
    celestials_installed = True
except RuntimeError:
    celestials_installed = False


@dataclass
class Entity:
    id: int

@dataclass
class SpaceRock:
    id: int
    name: str
    orbits_name: str
    is_moon: bool

@tag('library')
@skipIf(not celestials_installed, "Celestials Testpackage not installed into INSTALLED_APPS.")
class TestQueryBuilder(TestCase):

    def setUp(self):
        super().setUp()
        Stars.create_sol(context=self)

    def test_records(self):
        entities = Celestial.objects.filter(orbits__name='Sol', celestial_type__lte=4).records(Entity)
        self.assertEqual(len(entities), len(self.planets))

        # test whether we really return dataclass as result, even with first.
        self.assertIsInstance(entities.first(), Entity)

        # find moons. test whether i can use entities to do an SQL query. works because i have only one key.
        self.assertEqual(len(self.moons), Celestial.objects.filter(orbits__in=entities).count())
        
        # this is pretty much the same as
        self.assertEqual(len(self.moons), len(Celestial.objects.filter(
            orbits__in=Celestial.objects.filter(orbits__name='Sol', celestial_type__lte=4)).values_list('id', flat=True)))

    def test_lambda(self):

        # this tests whether our own celestial type or the celestial type of what we orbit is correct for being a moon. parameter is a dictionary.
        is_moon = lambda entry: True if 5 > (entry.get('celestial_type') or 0) > 1 and 5 > (entry.get('orbits_type') or 0) > 1 else False

        entities = Celestial.objects.records(SpaceRock,  # we want our output to be a SpaceRock dataclass.
                                    'celestial_type',  #  we include the key celestial_type into our query.
                                    id=Adjunct(None),  # we blank out id to test Adjunct working.
                                    orbits_name=F('orbits__name'),  # we set our custom orbits_name to a related field value
                                    orbits_type=F('orbits__celestial_type'),  # our lambda needs this data.
                                    is_moon=Lambda(is_moon))  # lambda over result

        self.assertEqual(len(entities), len(self.celestials))

        for idx, entity in enumerate(entities):
            dbdata = self.celestials[idx]
            model = Celestial.objects.filter(id=dbdata.id).first()
            self.assertEqual(entity.name, dbdata.name)
            self.assertIsNone(entity.id)
            self.assertEqual(entity.is_moon, model.is_moon)

    def test_callbacks(self):
        side_effect = lambda x:x
        callback_one = mock.Mock(side_effect=side_effect)
        callback_two = mock.Mock(side_effect=side_effect)

        entities = Celestial.objects.all().records(Entity, callback_one=Callback(callback_one), callback_two=Callback(callback_two))

        self.assertEqual(len(entities), len(self.celestials))
        self.assertEqual(callback_one.call_count, len(self.celestials))
        self.assertEqual(callback_two.call_count, len(self.celestials))
    
    def test_double_value_technique(self):
        """
            Records open a new sort of technique for late calling details from dataclasses.
            imagine you have a dataclass called EntityIndex, which has only one field from the database: id.
            
            however it has custom fields, where you store a lambda expression.
            
            you could e.g. use it in a subquery, while still access the data.
            however this will hit the database twice if you evaluate the iterator yourself, as the lambda is not lazy.
        """
        
        planet_queryset = Celestial.objects.filter(orbits__name='Sol', celestial_type__lte=4)
        
        @dataclass
        class DetailedEntity:
            id: int
            name: str
        
        @dataclass
        class IndexEntity:
            id: int
            detail: DetailedEntity
        
        def get_details_exec(data):
            return Celestial.objects.filter(pk=data.get('id')).records(DetailedEntity).first()
        
        get_details = mock.Mock(side_effect=get_details_exec)
        
        # retrieves data per key only.
        my_planets = planet_queryset.records(IndexEntity, detail=Lambda(get_details))
        my_moons = Celestial.objects.filter(orbits__in=my_planets).records(IndexEntity, detail=Lambda(get_details)) # legal
        # django does not consume the iterator internally for subqueries:
        self.assertEqual(get_details.call_count, 0)
        # consume it ourselves...
        for planet in my_planets:
            self.assertIsNotNone(planet.detail.name)
        self.assertEqual(get_details.call_count, len(self.planets))
        # but...
        self.assertEqual(len(my_moons), len(self.moons))
        self.assertEqual(get_details.call_count, len(self.planets) + len(self.moons))
        